/* $Id: vthrottle.c,v 1.30 2004/06/04 02:01:51 jose Exp $ */
/*
 * Copyright (c) 2003-2005 Jose Nazario <jose@monkey.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Jose Nazario.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <sys/param.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <err.h>
#include <errno.h>
#include <pthread.h>
#include <signal.h>
#include <stdarg.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <sysexits.h>
#include <syslog.h>
#include <time.h>
#include <unistd.h>
#include <util.h>

#include "libmilter/mfapi.h"
#include "sys/queue.h"

TAILQ_HEAD(tailhead, entry) head;
struct tailhead *headp;

struct entry {
	char            	hostname[256];	/* host's name */
	char            	mail_from[256]; /* account name */
	char			helo[256]; 	/* from HELO/EHLO command */ 
	char			hostip[NI_MAXHOST];	/* Inbound IP Address */
	time_t          	time_seen;	/* last time seen, seconds */ 
	u_int			interval;	/* expected time interval */
	u_int			throttled;	/* times blocked */
	u_int			flags;		/* info flags */
#define V_WHITELIST		0x001
#define V_THROTTLED		0x002
#define V_CONNECT		0x004
#define V_HELO			0x008
#define V_EFROM			0x010
	TAILQ_ENTRY(entry)	entries;	/* next entry */
};

struct hostinfo {
	char	hostname[256];
	char	hostip[NI_MAXHOST];
	char    mail_from[256]; /* account name */
	char	helo[256]; 	/* from HELO/EHLO command */ 
	int	flags;
	struct entry *e;
};
#define HIPRIV        ((struct hostinfo *) smfi_getpriv(ctx))

/* defaults */
int			interval = 60, expire = 0, nentries = 0, 
			maxentries = 8192, reclaimed = 0, pruned = 0,
			hentries = 0, debug = 0, throttled = 0, unthrottled = 0;
static pthread_mutex_t  le_mutex = PTHREAD_MUTEX_INITIALIZER;

#define SYSLOG_DATA_VTHROTTLE {-1, 0, 0, LOG_PID | LOG_NDELAY, NULL, LOG_MAIL, 0xff}

#define infonp(X,Y) \
	asprintf(&info,"remote=%s([%s]), HELO=%s, efrom=<%s>%s", \
		(X->flags & V_WHITELIST) && ! X->hostname[0] ? Y->hostname : X->hostname, \
		(X->flags & V_WHITELIST) ? Y->hostip : X->hostip, \
		X->helo, \
		X->mail_from, \
		(X->flags & V_WHITELIST) ? ", flags='WHITE'" : "")

char *fmtentry(struct entry *);
char *
fmtentry(struct entry *np)
{
	char *info, *ret, *type;
	infonp(np,np);
	switch(np->flags & (V_CONNECT|V_HELO|V_EFROM)) {
	case V_CONNECT:
		type = "host";
		break;
	case V_EFROM:
		type = "efrom";
		break;
	case V_HELO:
		type = "helo";
		break;
	default:
		type = "unknown";
		break;
	}
	asprintf(&ret, 
	    "%s: %s, expected interval: %d, last seen: %d, throttled %d times",
	    type, info, np->interval, np->time_seen, np->throttled);
	free(info);
	return ret;
}

void
dumpstats(int sig) 
{
	FILE 	      	       *statfile;
	time_t			now;
	static struct entry    *np;

	if (SIGUSR1 != sig)
		return;

	now = time(NULL);

	if (pthread_mutex_lock(&le_mutex) != 0) {
		warn("dumpstats cannot lock, bleh\n");
		return;		/* XXX */
	}

	if ((statfile = fopen("/var/tmp/vthrottle.stats", "a")) == NULL) {
		warn("statfile error in dumpstats.");
		return;
	}

	fprintf(statfile, "\nvthrottle statistics (%d) throttled: %d, unthrottled: %d,\n", now, throttled, unthrottled);
	fprintf(statfile, "cache stats: current size: %d, high size: %d, max: %d, reclaimed: %d\n", nentries, hentries, maxentries, reclaimed);
	fprintf(statfile, "pruned %d expired entries\n", pruned);

	/* print info for each entry */
	TAILQ_FOREACH(np, &head, entries) {
		char *loginfo;
		loginfo = fmtentry(np);
		fprintf(statfile, "%s\n", loginfo);
		free(loginfo);
	}

	fclose(statfile);
	pthread_mutex_unlock(&le_mutex);
	return;
}

/* 
 *  builds a whitelist of hosts and addresses that can send unlimited.
 *  returns 0 on success, -1 on failure.
 */

static int
parse_whitelist(char *filename)
{
	FILE			*file;
	static struct entry	*n2;
	char			*input;
	struct syslog_data	 sdata = SYSLOG_DATA_VTHROTTLE;

	if ((file = fopen(filename, "r")) == NULL) {
		warn("Failed to open whitelist file");
		return(-1);
	}

	for (; (input = fparseln(file, NULL, NULL, NULL, 0)); free(input)) {
		int	i = 0;
		char   *next, *entry[3];

		if (input == NULL)
			break;

		next = input;
		for (; i < 3 && (entry[i] = strsep(&next, " \t")) != NULL;)
			if (*entry[i] != '\0')
				i++;

		/* do we have a type/value/interval triplet? */
		if (i >= 2) {
			int match = 0;
			n2 = calloc(1,sizeof(struct entry));
			n2->time_seen = 0;
			n2->flags |= V_WHITELIST;
			if (strncasecmp(entry[0], "helo", 4) == 0) {
				n2->flags |= V_HELO;
				match = 1;
				snprintf(n2->hostname, 256, "%s", "");
				snprintf(n2->mail_from, 256, "%s", "");
				snprintf(n2->helo, 256, "%s", entry[1]);
			}
			if (strncasecmp(entry[0], "host", 4) == 0) {
				match = 1;
				n2->flags |= V_CONNECT;
				snprintf(n2->hostname, 256, "%s", entry[1]);
				snprintf(n2->mail_from, 256, "%s", "");
				snprintf(n2->helo, 256, "%s", "");
			}
			if (strncasecmp(entry[0], "efrom", 5) == 0)  {
				match = 1;
				n2->flags |= V_EFROM;
				snprintf(n2->hostname, 256, "%s", "");
				snprintf(n2->mail_from, 256, "%s", entry[1]);
				snprintf(n2->helo, 256, "%s", "");
			}
			if (match == 0) {
				free(n2);
				syslog_r(LOG_WARNING, &sdata,
					"Invalid type in whitelist: %s\n",
					entry[0]);
				continue;
			}
			n2->interval = atoi(entry[2]);
			TAILQ_INSERT_HEAD(&head, n2, entries);
			if (debug || 1) { /* XXX for temporary, always print */
				char *info = fmtentry(n2);
				syslog_r(LOG_WARNING, &sdata,
					"Adding to whitelist: %s\n",
					info);
				free(info);
			}
			nentries++;
			if (nentries > hentries)
				hentries = nentries;
			if (nentries > maxentries) {
				fprintf(stderr, "too many whitelist entries (more than maxentries), stopped parsing whitelist\n");
				break;
			}
		}
	}

	fclose(file);
	return(0);
}

/* 
 * removes outdated entries, based on the expiry timeout set by the user.
 * defaults to never expiring (is set to 0).
 */

void
prune(struct entry *np)
{
	char		       *description;
	struct syslog_data 	sdata = SYSLOG_DATA_VTHROTTLE;
	char *info;
	time_t now;

	if (!expire)
		return;

	if (!np)
		return;

	now = time(NULL);

	/* not too old */
	if ((now - np->time_seen) <= expire)
		return;

	/* don't prnue whitelist entries */
	if (np->flags & V_WHITELIST)
		return;

	/* remove stale entry */
	TAILQ_REMOVE(&head, np, entries);
	nentries--;
	pruned++;

	infonp(np,np);
	
	if (strlen(np->helo) > 1)
		asprintf(&description, "%s type=HELO,", info);
	else if (strlen(np->mail_from) > 1)
		asprintf(&description, "%s type=efrom,", info);
	else if (strlen(np->hostname) > 1)
		asprintf(&description, "%s type=remote,", info);	
	else 
		asprintf(&description, "unknown");

	if (debug) {
		syslog_r(LOG_WARNING, &sdata,
		"%s removing expired entry, age %d is older than %d\n",
			description, now - np->time_seen, expire);
	}

	free(description);
	free(info);

	free(np);
}

/* 
 * check the size of the list, remove the tail entries if it's too
 * long. skip the whitelist entries, remove the previous entry. leave
 * the whitelist entry to dangle at the bottom.
 */

void
check_lru_size(int num, struct entry *np)
{
	char		       *description;
	static struct entry    *n2;
	struct syslog_data      sdata = SYSLOG_DATA_VTHROTTLE;

	if (num > maxentries) {
		char *info;
		/* try again on a whitelist entry */
		if (np->flags & V_WHITELIST) {
			n2 = TAILQ_PREV(np, tailhead, entries);
			check_lru_size(num, n2);
			/* don't prune whitelist entries */
			return;
		}
		/* remove stale entry */
		TAILQ_REMOVE(&head, np, entries);

		infonp(np,np);
		if (strlen(np->helo) > 1)
			asprintf(&description, "%s type=HELO,", info);
		else if (strlen(np->mail_from) > 1)
			asprintf(&description, "%s type=efrom,", info);
		else if (strlen(np->hostname) > 1)
			asprintf(&description, "%s type=remote,", info);	
		else
			asprintf(&description, "unknown");

		syslog_r(LOG_WARNING, &sdata,
			"%s removing expired entry, list size exceeded (%d > %d) \n",
			description, nentries, maxentries);

		nentries--;
		reclaimed++;
		free(description);
		free(info);
		free(np);
	}
}

int
decide(struct entry *np, struct hostinfo *priv, char *type)
{
	static time_t		now;
	char *info;
	struct syslog_data sdata = SYSLOG_DATA_VTHROTTLE;

	infonp(np,priv);

	now = time(NULL);

	/* no more than one connection per interval ... */
	if ((now - np->time_seen) < (np->interval)) {
		syslog_r(LOG_WARNING, &sdata,
		    "%s throttling %s, waited=%ds, forced=%ds, seen = %d, now = %d\n", info, type, now - np->time_seen, np->interval, np->time_seen, now);
			np->throttled += 1;
		TAILQ_REMOVE(&head, np, entries);
		TAILQ_INSERT_HEAD(&head, np, entries);
		priv->flags |= V_THROTTLED;
		pthread_mutex_unlock(&le_mutex);
		free(info);
		return SMFIS_TEMPFAIL;
	}

	/* else */

	np->time_seen = now;
	TAILQ_REMOVE(&head, np, entries);
	TAILQ_INSERT_HEAD(&head, np, entries);
	syslog_r(LOG_WARNING, &sdata,
		"%s passing %s, seen = %d\n", info, type, np->time_seen);
	pthread_mutex_unlock(&le_mutex);
	free(info);
	return SMFIS_CONTINUE;
}

sfsistat
w_connect(SMFICTX *ctx, char *hostname, _SOCK_ADDR *hostaddr)
{
	static struct entry    *np, *n2, *nxt;
	struct syslog_data      sdata = SYSLOG_DATA_VTHROTTLE;
	struct hostinfo *priv;
	char host[NI_MAXHOST];

	if (smfi_setreply(ctx, "451", "4.7.1", "Please try again later") == MI_FAILURE) { 
		struct syslog_data sdata = SYSLOG_DATA_VTHROTTLE;
		syslog_r(LOG_WARNING, &sdata, "w_connect, cannot set reply to 451");
	}

	/* allocate some private memory */
	priv = calloc(1,sizeof(struct hostinfo));
	if (priv == NULL) {
		return SMFIS_TEMPFAIL;
	}
	memset(priv, '\0', sizeof(struct hostinfo));

	/* save the private data */
	smfi_setpriv(ctx, priv);

	if (pthread_mutex_lock(&le_mutex) != 0) {
		struct syslog_data sdata = SYSLOG_DATA_VTHROTTLE;
		syslog_r(LOG_WARNING, &sdata, "w_connect cannot lock mutex");
		priv->flags |= V_THROTTLED;
		return SMFIS_TEMPFAIL;    /* XXX */
	}

	/* has this entry been seen before? */
	for (np = TAILQ_FIRST(&head); np != TAILQ_END(&head); np = nxt) {
		nxt = TAILQ_NEXT(np, entries);
		if ((np->hostname != NULL) && (np->flags & V_CONNECT) &&
		    (strncasecmp(hostname, np->hostname, 
		     strlen(hostname)) == 0)) {
			if (np->flags & V_WHITELIST) {
				priv->flags |= V_WHITELIST;
			}
			/* unlocks mutex before returning */
			return decide(np, priv, "remote");
		} else
			prune(np);
	}

	/* new entry */
	n2 = calloc(1,sizeof(struct entry));
	if (n2 == NULL) 
		return SMFIS_CONTINUE; 		/* XXX - don't fail closed */
	n2->time_seen = time(NULL);
	n2->flags = V_CONNECT;
	n2->throttled = 0;

	snprintf(n2->hostname, 256, "%s", hostname);
	snprintf(n2->helo, 256, "%s", "");
	snprintf(n2->mail_from, 256, "%s", "");
	snprintf(priv->hostname, 256, "%s", hostname);
	snprintf(priv->helo, 256, "%s", "");
	snprintf(priv->mail_from, 256, "%s", "");
	strlcpy(host, "unknown", sizeof(host));
	if(hostaddr != NULL) {
	switch (hostaddr->sa_family) {
	case AF_INET: {
		struct sockaddr_in *sin = (struct sockaddr_in *)hostaddr;
		if (inet_ntop(AF_INET, &sin->sin_addr.s_addr, host,
			sizeof(host)) == NULL) {
			syslog_r(LOG_WARNING, &sdata,
				"w_connect: inet_ntop: %s\n", strerror(errno));
		}
		if (host == NULL) {
			syslog_r(LOG_WARNING, &sdata,
				"w_connect: inet_ntop(host): %s\n", strerror(errno));
		}
		break;
	}
	case AF_INET6: {
		struct sockaddr_in6 *sin6 = (struct sockaddr_in6 *)hostaddr;
		if (inet_ntop(AF_INET6, &sin6->sin6_addr, host,
			sizeof(host)) == NULL) {
			syslog_r(LOG_WARNING, &sdata,
				"w_connect: inet_ntop: %s\n", strerror(errno));
		}
		if (host == NULL) {
			syslog_r(LOG_WARNING, &sdata,
				"w_connect: inet_ntop(host): %s\n", strerror(errno));
		}
		break;
	}
	}
	}
	strlcpy(n2->hostip, host, sizeof(n2->hostip));
	strlcpy(priv->hostip, host, sizeof(priv->hostip));
		
	n2->interval = interval;
	TAILQ_INSERT_HEAD(&head, n2, entries);
	nentries++;
	if (nentries > hentries)
		hentries = nentries;
	check_lru_size(nentries, TAILQ_LAST(&head, tailhead));

	pthread_mutex_unlock(&le_mutex);
	return SMFIS_CONTINUE;
}

sfsistat
w_helo(SMFICTX * ctx, char *helohost)
{
	static struct entry    *np, *n2, *nxt;
	struct hostinfo *priv = HIPRIV;

	if (priv->flags & V_WHITELIST)
		return SMFIS_CONTINUE;

	if (smfi_setreply(ctx, "451", "4.7.1", "Please try again later") == MI_FAILURE) { 
		struct syslog_data sdata = SYSLOG_DATA_VTHROTTLE;
		syslog_r(LOG_WARNING, &sdata, "w_helo, cannot set reply to 451");
	}
	if (priv->flags & V_THROTTLED)
		return SMFIS_TEMPFAIL;

	if (pthread_mutex_lock(&le_mutex) != 0) {
		struct syslog_data sdata = SYSLOG_DATA_VTHROTTLE;
		syslog_r(LOG_WARNING, &sdata, "w_helo cannot lock mutex");
		priv->flags |= V_THROTTLED;
		return SMFIS_TEMPFAIL;    /* XXX */
	}

	/* has this entry been seen before? */
	for (np = TAILQ_FIRST(&head); np != TAILQ_END(&head); np = nxt) {
		nxt = TAILQ_NEXT(np, entries);
		if ((np->helo != NULL) && (np->flags & V_HELO) &&
		    (strncasecmp(helohost, np->helo, 
		     strlen(helohost)) == 0)) {

			if (np->flags & V_WHITELIST) {
				priv->flags |= V_WHITELIST;
			}
			/* unlocks mutex before returning */
			return decide(np, priv, "HELO");
		} else
			prune(np);
	}

	/* new entry */
	n2 = calloc(1,sizeof(struct entry));
	if (n2 == NULL)
		return SMFIS_CONTINUE; 	/* XXX - but don't fail closed */
	n2->time_seen = time(NULL);
	n2->flags = V_HELO;
	n2->throttled = 0;

	snprintf(n2->helo, 256, "%s", helohost);
	snprintf(n2->hostname, 256, "%s", priv->hostname);
	snprintf(n2->hostip, NI_MAXHOST, "%s", priv->hostip);
	snprintf(n2->mail_from, 256, "%s", "");
	snprintf(priv->helo, 256, "%s", helohost);
	n2->interval = interval;
	TAILQ_INSERT_HEAD(&head, n2, entries);
	nentries++;
	if (nentries > hentries)
		hentries = nentries;
	check_lru_size(nentries, TAILQ_LAST(&head, tailhead));

	pthread_mutex_unlock(&le_mutex);
	return SMFIS_CONTINUE;
}

sfsistat
w_envfrom(SMFICTX * ctx, char **argv)
{
	static char            *mailaddr;
	static struct entry    *np, *n2, *nxt;
	struct hostinfo *priv = HIPRIV;

	if (priv->flags & V_WHITELIST)
		return SMFIS_CONTINUE;

	if (smfi_setreply(ctx, "451", "4.7.1", "Please try again later") == MI_FAILURE) { 
		struct syslog_data sdata = SYSLOG_DATA_VTHROTTLE;
		syslog_r(LOG_WARNING, &sdata, "w_helo, cannot set reply to 451");
	}
	if (priv->flags & V_THROTTLED)
		return SMFIS_TEMPFAIL;

	if (pthread_mutex_lock(&le_mutex) != 0) {
		warn("w_from cannot lock, bleh\n");
		priv->flags |= V_THROTTLED;
		return SMFIS_TEMPFAIL;    /* XXX */
	}

	mailaddr = smfi_getsymval(ctx, "{mail_addr}");

	/* walk the list first ... */
	for (np = TAILQ_FIRST(&head); np != TAILQ_END(&head); np = nxt) {
		nxt = TAILQ_NEXT(np, entries);
		if ((np->flags & V_EFROM) && strncasecmp(mailaddr, np->mail_from,
		    sizeof(np->mail_from)) == 0) {

			if (np->flags & V_WHITELIST) {
				priv->flags |= V_WHITELIST;
			}
			/* unlocks mutex before returning */
			return decide(np, priv, "efrom");

		} else
			prune(np);
	}

	/* new entry */
	n2 = calloc(1,sizeof(struct entry));
	if (n2 == NULL)
		return SMFIS_CONTINUE;	/* XXX - but don't fail closed */
	n2->flags = V_EFROM;
	n2->throttled = 0;

	snprintf(n2->mail_from, 256, "%s", mailaddr);
	snprintf(n2->helo, 256, "%s", "");
	snprintf(n2->hostname, 256, "%s", priv->hostname);
	snprintf(n2->hostip, NI_MAXHOST, "%s", priv->hostip);
	snprintf(priv->mail_from, 256, "%s", mailaddr);
	n2->time_seen = time(NULL);
	n2->interval = interval;
	TAILQ_INSERT_HEAD(&head, n2, entries);
	nentries++;
	if (nentries > hentries)
		hentries = nentries;
	check_lru_size(nentries, TAILQ_LAST(&head, tailhead));

	pthread_mutex_unlock(&le_mutex);
	return SMFIS_CONTINUE;
}

sfsistat
w_close(SMFICTX *ctx)
{
	struct hostinfo *priv = HIPRIV;

	if (priv == NULL)
		return SMFIS_CONTINUE;

	if (priv->flags & V_THROTTLED) {
		throttled++;
	} else {
		unthrottled++;
	}
	free(priv);
	smfi_setpriv(ctx, NULL);
	return SMFIS_CONTINUE;
}

struct smfiDesc smfilter =
{
	"vthrottle",	/* filter name */
	SMFI_VERSION,	/* version code -- do not change */
	SMFIF_ADDHDRS,
	w_connect,	/* connection info filter */
	w_helo,		/* SMTP HELO command filter */
	w_envfrom,	/* envelope sender filter */
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	w_close,	/* connection cleanup */
};

static void
usage(char *prog)
{
	fprintf(stderr, 
	    "usage: %s -p sockaddr [-i interval] [-m maxentries] [-w whitelist] [-x expire]\n", prog);
	exit(1);
}

/*
 * an implementation of m. williamson's rate throttling mechanism for
 * mail bourne worms and viruses. basically we keep track of what hosts
 * and what senders have been sending mail. if they attempt to send mail
 * from one machine or using one address more than once in any interval
 * we back them off by sending a (temporary) failure. well behaved hosts
 * are expected to not trip this detection/throttle ...
 *
 * we keep the observations in a singly linked list as an LRU, so we
 * update the head of the LRU when we update an entry. keeps popular 
 * entries near the head of the list so they're more quickly found. the
 * list is maxentries long (default of 8192 entries)
 *
 * references:
 * Williamson, M. HPL-2002-172
 *      HP Labs Tech Report HPL-2002-172
 * 	http://www.hpl.hp.com/techreports/2002/HPL-2002-172.html
 * Twycross, J., and Williamson, M.
 * 	in Proc. 12th USENIX Security Symposium, pp 285-294
 *	http://www.usenix.org/events/sec03/tech/twycross.html
 */

int
main(int argc, char **argv)
{
	int            		c, setconn = 0, retval;
	extern char    	       *optarg;
	struct syslog_data 	sdata = SYSLOG_DATA_VTHROTTLE;

	TAILQ_INIT(&head); 

	/* Earlier than normal so we can log whitelist parsing */
	openlog_r("vthrottle", LOG_PID | LOG_NDELAY, LOG_MAIL, &sdata);

	while ((c = getopt(argc, argv, "i:m:p:w:x:h")) != -1) {
		switch (c) {
		case 'i':
			interval = atoi(optarg);
			break;
		case 'm':
			if (atoi(optarg) < 64) 
				fprintf(stderr, "WARNING: a very small list size can defeat detection, continuing anyhow\n");
			maxentries = atoi(optarg);
			break;
		case 'p':
			/* XXX - need to check for this file first */
			if (smfi_setconn(optarg) == MI_FAILURE)
				err(1, "socket connect failure.");
			if (strncmp(optarg, "unix:", 5) == 0)
				unlink(optarg + 5);
			if (strncmp(optarg, "local:", 6) == 0)
				unlink(optarg + 6);
			setconn = 1;
			break;
		case 'w':
			if (parse_whitelist(optarg) < 0)
				err(1, "whitelist file malformed.");
			break;
		case 'x':
			expire = atoi(optarg);
			if (expire < 1)
				err(1, "expiration interval must be greater than 0.");
			if (expire < interval)
				err(1, "expiration interval must be greater than the default throttle interval.");
			break;
		case 'h':
		default:
			usage(argv[0]);
			break;
		}
	}

	if (!setconn) 
		usage(argv[0]);

	/* XXX threadsafe? */
	signal(SIGUSR1, dumpstats);

	if (smfi_register(smfilter) == MI_FAILURE) 
		err(1, "initialization failure.");

	retval = smfi_main();
	closelog();
	pthread_mutex_destroy(&le_mutex);
	return retval;
}
